/*
  stepper.c - stepper motor driver: executes motion plans using stepper motors
  Part of Grbl

  Copyright (c) 2009-2011 Simen Svale Skogsrud

  Grbl is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Grbl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Grbl.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "Marlin.h"
#include "stepper.h"
#include "planner.h"
#include "temperature.h"
#include "language.h"
#include "cardreader.h"
#include "speed_lookuptable.h"
#if defined(DigiPotssPin) && DigiPotssPin > -1
#include <SPI.h>
#endif


block_t *current_block; 
static unsigned char out_bits;
static long counter_x,   
            counter_y,
            counter_z,
            counter_e;
volatile static unsigned long step_events_completed; 
#ifdef ADVANCE
  static long advance_rate, advance, final_advance = 0;
  static long old_advance = 0;
  static long e_steps[3];
#endif
static long acceleration_time, deceleration_time;
static unsigned short acc_step_rate; 
static char step_loops;
static unsigned short OCR1A_nominal;
static unsigned short step_loops_nominal;
volatile long endstops_trigsteps[3]={0,0,0};
volatile long endstops_stepsTotal,endstops_stepsDone;
static volatile bool endstop_x_hit=false;
static volatile bool endstop_y_hit=false;
static volatile bool endstop_z_hit=false;
#ifdef ABORT_ON_ENDSTOP_HIT_FEATURE_ENABLED
bool abort_on_endstop_hit = false;
#endif

static bool old_x_min_endstop=false;
static bool old_x_max_endstop=false;
static bool old_y_min_endstop=false;
static bool old_y_max_endstop=false;
static bool old_z_min_endstop=false;
static bool old_z_max_endstop=false;
static bool check_endstops = true;
volatile long count_position[NumAxis] = { 0, 0, 0, 0};
volatile signed char count_direction[NumAxis] = { 1, 1, 1, 1};
#define CHECK_ENDSTOPS  if(check_endstops)
#define MultiU16X8toH16(intRes, charIn1, intIn2) \
asm volatile ( \
"clr r26 \n\t" \
"mul %A1, %B2 \n\t" \
"movw %A0, r0 \n\t" \
"mul %A1, %A2 \n\t" \
"add %A0, r1 \n\t" \
"adc %B0, r26 \n\t" \
"lsr r0 \n\t" \
"adc %A0, r26 \n\t" \
"adc %B0, r26 \n\t" \
"clr r1 \n\t" \
: \
"=&r" (intRes) \
: \
"d" (charIn1), \
"d" (intIn2) \
: \
"r26" \
)
#define MultiU24X24toH16(intRes, longIn1, longIn2) \
asm volatile ( \
"clr r26 \n\t" \
"mul %A1, %B2 \n\t" \
"mov r27, r1 \n\t" \
"mul %B1, %C2 \n\t" \
"movw %A0, r0 \n\t" \
"mul %C1, %C2 \n\t" \
"add %B0, r0 \n\t" \
"mul %C1, %B2 \n\t" \
"add %A0, r0 \n\t" \
"adc %B0, r1 \n\t" \
"mul %A1, %C2 \n\t" \
"add r27, r0 \n\t" \
"adc %A0, r1 \n\t" \
"adc %B0, r26 \n\t" \
"mul %B1, %B2 \n\t" \
"add r27, r0 \n\t" \
"adc %A0, r1 \n\t" \
"adc %B0, r26 \n\t" \
"mul %C1, %A2 \n\t" \
"add r27, r0 \n\t" \
"adc %A0, r1 \n\t" \
"adc %B0, r26 \n\t" \
"mul %B1, %A2 \n\t" \
"add r27, r1 \n\t" \
"adc %A0, r26 \n\t" \
"adc %B0, r26 \n\t" \
"lsr r27 \n\t" \
"adc %A0, r26 \n\t" \
"adc %B0, r26 \n\t" \
"clr r1 \n\t" \
: \
"=&r" (intRes) \
: \
"d" (longIn1), \
"d" (longIn2) \
: \
"r26" , "r27" \
)

#define ENABLE_STEPPER_DRIVER_INTERRUPT()  TIMSK1 |= (1<<OCIE1A)
#define DISABLE_STEPPER_DRIVER_INTERRUPT() TIMSK1 &= ~(1<<OCIE1A)


void checkHitEndstops()
{
 if( endstop_x_hit || endstop_y_hit || endstop_z_hit) 
 {
   SERIAL_ECHO_START;
   SERIAL_ECHOPGM(MSG_ENDSTOPS_HIT);
   if(endstop_x_hit) 
   {
     SERIAL_ECHOPAIR(" X:",(float)endstops_trigsteps[X_AXIS]/axis_steps_per_unit[X_AXIS]);
   }
   if(endstop_y_hit) 
   {
     SERIAL_ECHOPAIR(" Y:",(float)endstops_trigsteps[Y_AXIS]/axis_steps_per_unit[Y_AXIS]);
   }
   if(endstop_z_hit) 
   {
     SERIAL_ECHOPAIR(" Z:",(float)endstops_trigsteps[Z_AXIS]/axis_steps_per_unit[Z_AXIS]);
   }
   SERIAL_ECHOLN("");
   endstop_x_hit=false;
   endstop_y_hit=false;
   endstop_z_hit=false;
#ifdef ABORT_ON_ENDSTOP_HIT_FEATURE_ENABLED
   if (abort_on_endstop_hit)
   {
     card.sdprinting = false;
     card.closefile();
     quickStop();
     setTargetHotend0(0);
     setTargetHotend1(0);
     setTargetHotend2(0);
     setTargetHotend3(0);
     setTargetHotend4(0);
   }
#endif
 }
}

void endstops_hit_on_purpose()
{
  endstop_x_hit=false;
  endstop_y_hit=false;
  endstop_z_hit=false;
}

void enable_endstops(bool check)
{
  check_endstops = check;
}

void st_wake_up() 
{
  ENABLE_STEPPER_DRIVER_INTERRUPT();
}

void step_wait(){
    for(int8_t i=0; i < 6; i++){
    }
}

FORCE_INLINE unsigned short calc_timer(unsigned short step_rate) 
{
  unsigned short timer;
  if(step_rate > MAX_STEP_FREQUENCY) step_rate = MAX_STEP_FREQUENCY;

  if(step_rate > 20000) 
  { 
    step_rate = (step_rate >> 2)&0x3fff;
    step_loops = 4;
  }
  else if(step_rate > 10000) 
  { 
    step_rate = (step_rate >> 1)&0x7fff;
    step_loops = 2;
  }
  else {
    step_loops = 1;
  }

  if(step_rate < (F_CPU/500000)) step_rate = (F_CPU/500000);
  step_rate -= (F_CPU/500000); // Correct for minimal speed
  if(step_rate >= (8*256))
  { // higher step rate
    unsigned short table_address = (unsigned short)&speed_lookuptable_fast[(unsigned char)(step_rate>>8)][0];
    unsigned char tmp_step_rate = (step_rate & 0x00ff);
    unsigned short gain = (unsigned short)pgm_read_word_near(table_address+2);
    MultiU16X8toH16(timer, tmp_step_rate, gain);
    timer = (unsigned short)pgm_read_word_near(table_address) - timer;
  }
  else 
  { 
    unsigned short table_address = (unsigned short)&speed_lookuptable_slow[0][0];
    table_address += ((step_rate)>>1) & 0xfffc;
    timer = (unsigned short)pgm_read_word_near(table_address);
    timer -= (((unsigned short)pgm_read_word_near(table_address+2) * (unsigned char)(step_rate & 0x0007))>>3);
  }
  if(timer < 100) { timer = 100; MYSERIAL.print(MSG_STEPPER_TOO_HIGH); MYSERIAL.println(step_rate); }//(20kHz this should never happen)
  return timer;
}

FORCE_INLINE void trapezoid_generator_reset() 
{
  #ifdef ADVANCE
    advance = current_block->initial_advance;
    final_advance = current_block->final_advance;
    e_steps[current_block->active_extruder] += ((advance >>8) - old_advance);
    old_advance = advance >>8;
  #endif
  deceleration_time = 0;
  OCR1A_nominal = calc_timer(current_block->nominal_rate);
  step_loops_nominal = step_loops;
  acc_step_rate = current_block->initial_rate;
  acceleration_time = calc_timer(acc_step_rate);
  OCR1A = acceleration_time;
  }

ISR(TIMER1_COMPA_vect)
{
  if (current_block == NULL) {
    current_block = plan_get_current_block();
    if (current_block != NULL) {
      current_block->busy = true;
      trapezoid_generator_reset();
      counter_x = -(current_block->step_event_count >> 1);
      counter_y = counter_x;
      counter_z = counter_x;
      counter_e = counter_x;
      step_events_completed = 0;

      #ifdef Z_LATE_ENABLE
        if(current_block->steps_z > 0) {
          enable_z();
          OCR1A = 2000; //1ms wait
          return;
        }
      #endif
    }
    else {
        OCR1A=2000;
    }
  }

  if (current_block != NULL) {
    out_bits = current_block->direction_bits;

    if((out_bits & (1<<X_AXIS))!=0){
              WRITE(XdirPin, InvertXdir);
      count_direction[X_AXIS]=-1;
    }
    else{
        WRITE(XdirPin, !InvertXdir);      
      count_direction[X_AXIS]=1;
    }
    if((out_bits & (1<<Y_AXIS))!=0){
      WRITE(YdirPin, InvertYdir);
      count_direction[Y_AXIS]=-1;
    }
    else{
      WRITE(YdirPin, !InvertYdir);
      count_direction[Y_AXIS]=1;
    }

    #ifndef COREXY
    if ((out_bits & (1<<X_AXIS)) != 0) {
    #else
    if ((((out_bits & (1<<X_AXIS)) != 0)&&(out_bits & (1<<Y_AXIS)) != 0)) {
    #endif
      CHECK_ENDSTOPS
      {   
        {
          #if defined(XminPin) && XminPin > -1
            bool x_min_endstop=(READ(XminPin) != XminEndstopInverting);
            if(x_min_endstop && old_x_min_endstop && (current_block->steps_x > 0)) {
              endstops_trigsteps[X_AXIS] = count_position[X_AXIS];
              endstop_x_hit=true;
              step_events_completed = current_block->step_event_count;
            }
            old_x_min_endstop = x_min_endstop;
          #endif
        }
      }
    }
    else { // +direction
      CHECK_ENDSTOPS
      {     
        {
          #if defined(XmaxPin) && XmaxPin > -1
            bool x_max_endstop=(READ(XmaxPin) != XmaxEndstopInverting);
            if(x_max_endstop && old_x_max_endstop && (current_block->steps_x > 0)){
              endstops_trigsteps[X_AXIS] = count_position[X_AXIS];
              endstop_x_hit=true;
              step_events_completed = current_block->step_event_count;
            }
            old_x_max_endstop = x_max_endstop;
          #endif
        }
      }
    }

    #ifndef COREXY
    if ((out_bits & (1<<Y_AXIS)) != 0) {   // -direction
    #else
    if ((((out_bits & (1<<X_AXIS)) != 0)&&(out_bits & (1<<Y_AXIS)) == 0)) {   // -Y occurs for -A and +B
    #endif
      CHECK_ENDSTOPS
      {
        #if defined(YminPin) && YminPin > -1
          bool y_min_endstop=(READ(YminPin) != YminEndstopInverting);
          if(y_min_endstop && old_y_min_endstop && (current_block->steps_y > 0)) {
            endstops_trigsteps[Y_AXIS] = count_position[Y_AXIS];
            endstop_y_hit=true;
            step_events_completed = current_block->step_event_count;
          }
          old_y_min_endstop = y_min_endstop;
        #endif
      }
    }
    else { // +direction
      CHECK_ENDSTOPS
      {
        #if defined(YmaxPin) && YmaxPin > -1
          bool y_max_endstop=(READ(YmaxPin) != YmaxEndstopInverting);
          if(y_max_endstop && old_y_max_endstop && (current_block->steps_y > 0)){
            endstops_trigsteps[Y_AXIS] = count_position[Y_AXIS];
            endstop_y_hit=true;
            step_events_completed = current_block->step_event_count;
          }
          old_y_max_endstop = y_max_endstop;
        #endif
      }
    }

    if ((out_bits & (1<<Z_AXIS)) != 0) {   // -direction
      WRITE(ZdirPin,InvertZdir);
      
      count_direction[Z_AXIS]=-1;
      CHECK_ENDSTOPS
      {
        #if defined(ZminPin) && ZminPin > -1
          bool z_min_endstop=(READ(ZminPin) != ZminEndstopInverting);
          if(z_min_endstop && old_z_min_endstop && (current_block->steps_z > 0)) {
            endstops_trigsteps[Z_AXIS] = count_position[Z_AXIS];
            endstop_z_hit=true;
            step_events_completed = current_block->step_event_count;
          }
          old_z_min_endstop = z_min_endstop;
        #endif
      }
    }
    else { // +direction
      WRITE(ZdirPin,!InvertZdir);
	  
      count_direction[Z_AXIS]=1;
      CHECK_ENDSTOPS
      {
        #if defined(ZmaxPin) && ZmaxPin > -1
          bool z_max_endstop=(READ(ZmaxPin) != ZmaxEndstopInverting);
          if(z_max_endstop && old_z_max_endstop && (current_block->steps_z > 0)) {
            endstops_trigsteps[Z_AXIS] = count_position[Z_AXIS];
            endstop_z_hit=true;
            step_events_completed = current_block->step_event_count;
          }
          old_z_max_endstop = z_max_endstop;
        #endif
      }
    }

    #ifndef ADVANCE
      if ((out_bits & (1<<E_AXIS)) != 0) {  // -direction
        REV_E_DIR();
        count_direction[E_AXIS]=-1;
      }
      else { // +direction
        NORM_E_DIR();
        count_direction[E_AXIS]=1;
      }
    #endif //!ADVANCE



    for(int8_t i=0; i < step_loops; i++) { // Take multiple steps per interrupt (For high speed moves)
      #ifndef AT90USB
      MSerial.checkRx(); // Check for serial chars.
      #endif

      #ifdef ADVANCE
      counter_e += current_block->steps_e;
      if (counter_e > 0) {
        counter_e -= current_block->step_event_count;
        if ((out_bits & (1<<E_AXIS)) != 0) { // - direction
          e_steps[current_block->active_extruder]--;
        }
        else {
          e_steps[current_block->active_extruder]++;
        }
      }
      #endif //ADVANCE

        counter_x += current_block->steps_x;
        if (counter_x > 0) 
		{        
          WRITE(XstepPin, !INVERT_XstepPin);
          counter_x -= current_block->step_event_count;
          count_position[X_AXIS]+=count_direction[X_AXIS];   
          WRITE(XstepPin, INVERT_XstepPin);
        }

        counter_y += current_block->steps_y;
        if (counter_y > 0) 
		{
          WRITE(YstepPin, !INVERT_YstepPin);
          counter_y -= current_block->step_event_count;
          count_position[Y_AXIS]+=count_direction[Y_AXIS];
          WRITE(YstepPin, INVERT_YstepPin);
        }

      counter_z += current_block->steps_z;
      if (counter_z > 0) 
	  {
        WRITE(ZstepPin, !INVERT_ZstepPin);
        
        #ifdef Z_DUAL_STEPPER_DRIVERS
          WRITE(Z2_STEP_PIN, !INVERT_ZstepPin);
        #endif

        counter_z -= current_block->step_event_count;
        count_position[Z_AXIS]+=count_direction[Z_AXIS];
        WRITE(ZstepPin, INVERT_ZstepPin);
        
        #ifdef Z_DUAL_STEPPER_DRIVERS
          WRITE(Z2_STEP_PIN, INVERT_ZstepPin);
        #endif
      }

      #ifndef ADVANCE
        counter_e += current_block->steps_e;
        if (counter_e > 0) 
		{
          WRITE_E_STEP(!INVERT_E_STEP_PIN);
          counter_e -= current_block->step_event_count;
          count_position[E_AXIS]+=count_direction[E_AXIS];
          WRITE_E_STEP(INVERT_E_STEP_PIN);
        }
      #endif 
      step_events_completed += 1;
      if(step_events_completed >= current_block->step_event_count) break;
    }

    unsigned short timer;
    unsigned short step_rate;
    if (step_events_completed <= (unsigned long int)current_block->accelerate_until) 
	{
      MultiU24X24toH16(acc_step_rate, acceleration_time, current_block->acceleration_rate);
      acc_step_rate += current_block->initial_rate;

      if(acc_step_rate > current_block->nominal_rate)
        acc_step_rate = current_block->nominal_rate;

      timer = calc_timer(acc_step_rate);
      OCR1A = timer;
      acceleration_time += timer;
      #ifdef ADVANCE
        for(int8_t i=0; i < step_loops; i++) 
		{
          advance += advance_rate;
        }
        e_steps[current_block->active_extruder] += ((advance >>8) - old_advance);
        old_advance = advance >>8;

      #endif
    }
    else if (step_events_completed > (unsigned long int)current_block->decelerate_after) {
      MultiU24X24toH16(step_rate, deceleration_time, current_block->acceleration_rate);

      if(step_rate > acc_step_rate) {
        step_rate = current_block->final_rate;
      }
      else {
        step_rate = acc_step_rate - step_rate;
      }

      if(step_rate < current_block->final_rate)
        step_rate = current_block->final_rate;

      timer = calc_timer(step_rate);
      OCR1A = timer;
      deceleration_time += timer;
      #ifdef ADVANCE
        for(int8_t i=0; i < step_loops; i++) 
		{
          advance -= advance_rate;
        }
        if(advance < final_advance) advance = final_advance;

        e_steps[current_block->active_extruder] += ((advance >>8) - old_advance);
        old_advance = advance >>8;
      #endif 
    }
    else {
      OCR1A = OCR1A_nominal;
      step_loops = step_loops_nominal;
    }

    if (step_events_completed >= current_block->step_event_count) {
      current_block = NULL;
      plan_discard_current_block();
    }
  }
}

#ifdef ADVANCE
  unsigned char old_OCR0A;
  ISR(TIMER0_COMPA_vect)
  {
    old_OCR0A += 52;
    OCR0A = old_OCR0A;
    for(unsigned char i=0; i<4;i++) 
	{
      if (e_steps[0] != 0) 
	  {
        WRITE(E0stepPin, INVERT_E_STEP_PIN);
        if (e_steps[0] < 0) 
		{
          WRITE(E0_
			  , InvertE0dir);
          e_steps[0]++;
          WRITE(E0stepPin, !INVERT_E_STEP_PIN);
        }
        else if (e_steps[0] > 0) 
		{
          WRITE(E0dirPin, !InvertE0dir);
          e_steps[0]--;
          WRITE(E0stepPin, !INVERT_E_STEP_PIN);
        }
      }
 #if Nozzles > 1
      if (e_steps[1] != 0) 
	  {
        WRITE(E1stepPin, INVERT_E_STEP_PIN);
        if (e_steps[1] < 0) 
		{
          WRITE(E1dirPin, InvertE1dir);
          e_steps[1]++;
          WRITE(E1stepPin, !INVERT_E_STEP_PIN);
        }
        else if (e_steps[1] > 0) 
		{
          WRITE(E1dirPin, !InvertE1dir);
          e_steps[1]--;
          WRITE(E1stepPin, !INVERT_E_STEP_PIN);
        }
      }
 #endif
 #if Nozzles > 2
      if (e_steps[2] != 0) 
	  {
        WRITE(E2stepPin, INVERT_E_STEP_PIN);
        if (e_steps[2] < 0) {
          WRITE(E2dirPin, InvertE2dir);
          e_steps[2]++;
          WRITE(E2stepPin, !INVERT_E_STEP_PIN);
        }
        else if (e_steps[2] > 0) 
		{
          WRITE(E2dirPin, !InvertE2dir);
          e_steps[2]--;
          WRITE(E2stepPin, !INVERT_E_STEP_PIN);
        }
      }
 #endif
 #if Nozzles > 3
      if (e_steps[3] != 0) 
	  {
        WRITE(E3stepPin, INVERT_E_STEP_PIN);
        if (e_steps[3] < 0) 
		{
          WRITE(E3dirPin, InvertE3dir);
          e_steps[3]++;
          WRITE(E3stepPin, !INVERT_E_STEP_PIN);
        }
        else if (e_steps[3] > 0) 
		{
          WRITE(E3dirPin, !InvertE3dir);
          e_steps[3]--;
          WRITE(E3stepPin, !INVERT_E_STEP_PIN);
        }
      }
 #endif
 #if Nozzles > 4
      if (e_steps[4] != 0) {
        WRITE(E4stepPin, INVERT_E_STEP_PIN);
        if (e_steps[4] < 0) {
          WRITE(E4dirPin, InvertE4dir);
          e_steps[4]++;
          WRITE(E4stepPin, !INVERT_E_STEP_PIN);
        }
        else if (e_steps[4] > 0) {
          WRITE(E4dirPin, !InvertE4dir);
          e_steps[4]--;
          WRITE(E4stepPin, !INVERT_E_STEP_PIN);
        }
      }
 #endif
    }
  }
#endif

void st_init()
{
  digipot_init(); 
  microstep_init(); 

  #if defined(XdirPin) && XdirPin > -1
    SET_OUTPUT(XdirPin);
  #endif
  #if defined(X2_DIR_PIN) && X2_DIR_PIN > -1
    SET_OUTPUT(X2_DIR_PIN);
  #endif
  #if defined(YdirPin) && YdirPin > -1
    SET_OUTPUT(YdirPin);
  #endif
  #if defined(ZdirPin) && ZdirPin > -1
    SET_OUTPUT(ZdirPin);

    #if defined(Z_DUAL_STEPPER_DRIVERS) && defined(Z2_DIR_PIN) && (Z2_DIR_PIN > -1)
      SET_OUTPUT(Z2_DIR_PIN);
    #endif
  #endif
  #if defined(E0dirPin) && E0dirPin > -1
    SET_OUTPUT(E0dirPin);
  #endif
  #if defined(E1dirPin) && (E1dirPin > -1)
    SET_OUTPUT(E1dirPin);
  #endif
  #if defined(E2dirPin) && (E2dirPin > -1)
    SET_OUTPUT(E2dirPin);
  #endif
  #if defined(E3dirPin) && (E3dirPin > -1)
    SET_OUTPUT(E3dirPin);
  #endif
  #if defined(E4dirPin) && (E4dirPin > -1)
    SET_OUTPUT(E4dirPin);
  #endif


  #if defined(XenablePin) && XenablePin > -1
    SET_OUTPUT(XenablePin);
    if(!XenableOn) WRITE(XenablePin,HIGH);
  #endif
  #if defined(X2_ENABLE_PIN) && X2_ENABLE_PIN > -1
    SET_OUTPUT(X2_ENABLE_PIN);
    if(!XenableOn) WRITE(X2_ENABLE_PIN,HIGH);
  #endif
  #if defined(YenablePin) && YenablePin > -1
    SET_OUTPUT(YenablePin);
    if(!YenableOn) WRITE(YenablePin,HIGH);
  #endif
  #if defined(ZenablePin) && ZenablePin > -1
    SET_OUTPUT(ZenablePin);
    if(!ZenableOn) WRITE(ZenablePin,HIGH);

    #if defined(Z_DUAL_STEPPER_DRIVERS) && defined(Z2_ENABLE_PIN) && (Z2_ENABLE_PIN > -1)
      SET_OUTPUT(Z2_ENABLE_PIN);
      if(!ZenableOn) WRITE(Z2_ENABLE_PIN,HIGH);
    #endif
  #endif
  #if defined(E0enablePin) && (E0enablePin > -1)
    SET_OUTPUT(E0enablePin);
    if(!EenableOn) WRITE(E0enablePin,HIGH);
  #endif
  #if defined(E1enablePin) && (E1enablePin > -1)
    SET_OUTPUT(E1enablePin);
    if(!EenableOn) WRITE(E1enablePin,HIGH);
  #endif
  #if defined(E2enablePin) && (E2enablePin > -1)
    SET_OUTPUT(E2enablePin);
    if(!EenableOn) WRITE(E2enablePin,HIGH);
  #endif
  #if defined(E3enablePin) && (E3enablePin > -1)
    SET_OUTPUT(E3enablePin);
    if(!EenableOn) WRITE(E3enablePin,HIGH);
  #endif
  #if defined(E4enablePin) && (E4enablePin > -1)
    SET_OUTPUT(E4enablePin);
    if(!EenableOn) WRITE(E4enablePin,HIGH);
  #endif

  #if defined(XminPin) && XminPin > -1
    SET_INPUT(XminPin);
    #ifdef EndstopPullupXmin
      WRITE(XminPin,HIGH);
    #endif
  #endif

  #if defined(YminPin) && YminPin > -1
    SET_INPUT(YminPin);
    #ifdef EndstopPullupYmin
      WRITE(YminPin,HIGH);
    #endif
  #endif

  #if defined(ZminPin) && ZminPin > -1
    SET_INPUT(ZminPin);
    #ifdef EndstopPullupZmin
      WRITE(ZminPin,HIGH);
    #endif
  #endif

  #if defined(XmaxPin) && XmaxPin > -1
    SET_INPUT(XmaxPin);
    #ifdef EndstopPullupXmax
      WRITE(XmaxPin,HIGH);
    #endif
  #endif

  #if defined(YmaxPin) && YmaxPin > -1
    SET_INPUT(YmaxPin);
    #ifdef EndstopPullupYmax
      WRITE(YmaxPin,HIGH);
    #endif
  #endif

  #if defined(ZmaxPin) && ZmaxPin > -1
    SET_INPUT(ZmaxPin);
    #ifdef EndstopPullupZmax
      WRITE(ZmaxPin,HIGH);
    #endif
  #endif

  #if defined(XstepPin) && (XstepPin > -1)
    SET_OUTPUT(XstepPin);
    WRITE(XstepPin,INVERT_XstepPin);
    Disable_X();
  #endif
  #if defined(X2_STEP_PIN) && (X2_STEP_PIN > -1)
    SET_OUTPUT(X2_STEP_PIN);
    WRITE(X2_STEP_PIN,INVERT_XstepPin);
    DisableX();
  #endif
  #if defined(YstepPin) && (YstepPin > -1)
    SET_OUTPUT(YstepPin);
    WRITE(YstepPin,INVERT_YstepPin);
    Disable_Y();
  #endif
  #if defined(ZstepPin) && (ZstepPin > -1)
    SET_OUTPUT(ZstepPin);
    WRITE(ZstepPin,INVERT_ZstepPin);
    #if defined(Z_DUAL_STEPPER_DRIVERS) && defined(Z2_STEP_PIN) && (Z2_STEP_PIN > -1)
      SET_OUTPUT(Z2_STEP_PIN);
      WRITE(Z2_STEP_PIN,INVERT_ZstepPin);
    #endif
    Disable_Z();
  #endif
  #if defined(E0stepPin) && (E0stepPin > -1)
    SET_OUTPUT(E0stepPin);
    WRITE(E0stepPin,INVERT_E_STEP_PIN);
    DisableE0();
  #endif
  #if defined(E1stepPin) && (E1stepPin > -1)
    SET_OUTPUT(E1stepPin);
    WRITE(E1stepPin,INVERT_E_STEP_PIN);
    DisableE1();
  #endif
  #if defined(E2stepPin) && (E2stepPin > -1)
    SET_OUTPUT(E2stepPin);
    WRITE(E2stepPin,INVERT_E_STEP_PIN);
    DisableE2();
  #endif
  #if defined(E3stepPin) && (E3stepPin > -1)
    SET_OUTPUT(E3stepPin);
    WRITE(E3stepPin,INVERT_E_STEP_PIN);
    DisableE3();
  #endif
  #if defined(E4stepPin) && (E4stepPin > -1)
    SET_OUTPUT(E4stepPin);
    WRITE(E4stepPin,INVERT_E_STEP_PIN);
    DisableE4();
  #endif

  TCCR1B &= ~(1<<WGM13);
  TCCR1B |=  (1<<WGM12);
  TCCR1A &= ~(1<<WGM11);
  TCCR1A &= ~(1<<WGM10);
  TCCR1A &= ~(3<<COM1A0);
  TCCR1A &= ~(3<<COM1B0);
  TCCR1B = (TCCR1B & ~(0x07<<CS10)) | (2<<CS10);

  OCR1A = 0x4000;
  TCNT1 = 0;
  ENABLE_STEPPER_DRIVER_INTERRUPT();

  #ifdef ADVANCE
  #if defined(TCCR0A) && defined(WGM01)
    TCCR0A &= ~(1<<WGM01);
    TCCR0A &= ~(1<<WGM00);
  #endif
    e_steps[0] = 0;
    e_steps[1] = 0;
    e_steps[2] = 0;
    e_steps[3] = 0;
    e_steps[4] = 0;
    TIMSK0 |= (1<<OCIE0A);
  #endif

  enable_endstops(true);
  sei();
}

void st_synchronize()
{
    while( blocks_queued()) 
	{
		manage_heater();
		manage_inactivity();
	}
}

void st_set_position(const long &x, const long &y, const long &z, const long &e)
{
  CRITICAL_SECTION_START;
  count_position[X_AXIS] = x;
  count_position[Y_AXIS] = y;
  count_position[Z_AXIS] = z;
  count_position[E_AXIS] = e;
  CRITICAL_SECTION_END;
}

void st_set_e_position(const long &e)
{
  CRITICAL_SECTION_START;
  count_position[E_AXIS] = e;
  CRITICAL_SECTION_END;
}

long st_get_position(uint8_t axis)
{
  long count_pos;
  CRITICAL_SECTION_START;
  count_pos = count_position[axis];
  CRITICAL_SECTION_END;
  return count_pos;
}

void finishAndDisableSteppers()
{
  st_synchronize();
  Disable_X();
  Disable_Y();
  Disable_Z();
  DisableE0();
  DisableE1();
  DisableE2();
  DisableE3();
  DisableE4();
}

void quickStop()
{
  DISABLE_STEPPER_DRIVER_INTERRUPT();
  while(blocks_queued())
    plan_discard_current_block();
  current_block = NULL;
  ENABLE_STEPPER_DRIVER_INTERRUPT();
}

void digitalPotWrite(int address, int value)
{
  #if defined(DigiPotssPin) && DigiPotssPin > -1
    digitalWrite(DigiPotssPin,LOW); 
    SPI.transfer(address); 
    SPI.transfer(value);
    digitalWrite(DigiPotssPin,HIGH); 
    //delay(10);
  #endif
}

void digipot_init() 
{
  #if defined(DigiPotssPin) && DigiPotssPin > -1
    const uint8_t digipot_motor_current[] = DIGIPOT_MOTOR_CURRENT;

    SPI.begin();
    pinMode(DigiPotssPin, OUTPUT);
    for(int i=0;i<=4;i++)
      digipot_current(i,digipot_motor_current[i]);
  #endif
}

void digipot_current(uint8_t driver, int current)
{
  #if defined(DigiPotssPin) && DigiPotssPin > -1
    const uint8_t digipot_ch[] = DIGIPOT_CHANNELS;
    digitalPotWrite(digipot_ch[driver], current);
  #endif
}

void microstep_init()
{
  #if defined(Xms1Pin) && Xms1Pin > -1
  const uint8_t microstep_modes[] = MICROSTEP_MODES;
  pinMode(Xms2Pin,OUTPUT);
  pinMode(Yms2Pin,OUTPUT);
  pinMode(Zms2Pin,OUTPUT);
  pinMode(E0ms2Pin,OUTPUT);
  pinMode(E1ms2Pin,OUTPUT);
  pinMode(E2ms2Pin,OUTPUT);
  pinMode(E3ms2Pin,OUTPUT);
  pinMode(E4ms2Pin,OUTPUT);
  for(int i=0;i<=7;i++) microstep_mode(i,microstep_modes[i]);
  #endif
}

void microstep_ms(uint8_t driver, int8_t ms1, int8_t ms2)
{
  if(ms1 > -1) switch(driver)
  {
    case 0: digitalWrite( Xms1Pin,ms1); break;
    case 1: digitalWrite( Yms1Pin,ms1); break;
    case 2: digitalWrite( Zms1Pin,ms1); break;
    case 3: digitalWrite(E0ms1Pin,ms1); break;
    case 4: digitalWrite(E1ms1Pin,ms1); break;
    case 5: digitalWrite(E2ms1Pin,ms1); break;
    case 6: digitalWrite(E3ms1Pin,ms1); break;
    case 7: digitalWrite(E4ms1Pin,ms1); break;
  }
  if(ms2 > -1) switch(driver)
  {
    case 0: digitalWrite( Xms2Pin,ms2); break;
    case 1: digitalWrite( Yms2Pin,ms2); break;
    case 2: digitalWrite( Zms2Pin,ms2); break;
    case 3: digitalWrite(E0ms2Pin,ms2); break;
    case 4: digitalWrite(E1ms2Pin,ms2); break;
    case 5: digitalWrite(E2ms2Pin,ms2); break;
    case 6: digitalWrite(E3ms2Pin,ms2); break;
    case 7: digitalWrite(E4ms2Pin,ms2); break;
  }
}

void microstep_mode(uint8_t driver, uint8_t stepping_mode)
{
  switch(stepping_mode)
  {
    case 1: microstep_ms(driver,MICROSTEP1); break;
    case 2: microstep_ms(driver,MICROSTEP2); break;
    case 4: microstep_ms(driver,MICROSTEP4); break;
    case 8: microstep_ms(driver,MICROSTEP8); break;
    case 16: microstep_ms(driver,MICROSTEP16); break;
    case 32: microstep_ms(driver,MICROSTEP32); break;
  }
}

void microstep_readings()
{
      SERIAL_PROTOCOLPGM("MS1,MS2 Pins\n");
      SERIAL_PROTOCOLPGM("X: ");
      SERIAL_PROTOCOL(   digitalRead(Xms1Pin));
      SERIAL_PROTOCOLLN( digitalRead(Xms2Pin));
      SERIAL_PROTOCOLPGM("Y: ");
      SERIAL_PROTOCOL(   digitalRead(Yms1Pin));
      SERIAL_PROTOCOLLN( digitalRead(Yms2Pin));
      SERIAL_PROTOCOLPGM("Z: ");
      SERIAL_PROTOCOL(   digitalRead(Zms1Pin));
      SERIAL_PROTOCOLLN( digitalRead(Zms2Pin));
      SERIAL_PROTOCOLPGM("E0: ");
      SERIAL_PROTOCOL(   digitalRead(E0ms1Pin));
      SERIAL_PROTOCOLLN( digitalRead(E0ms2Pin));
      SERIAL_PROTOCOLPGM("E1: ");
      SERIAL_PROTOCOL(   digitalRead(E1ms1Pin));
      SERIAL_PROTOCOLLN( digitalRead(E1ms2Pin));
      SERIAL_PROTOCOLPGM("E2: ");
      SERIAL_PROTOCOL(   digitalRead(E2ms1Pin));
      SERIAL_PROTOCOLLN( digitalRead(E2ms2Pin));
      SERIAL_PROTOCOLPGM("E3: ");
      SERIAL_PROTOCOL(   digitalRead(E3ms1Pin));
      SERIAL_PROTOCOLLN( digitalRead(E3ms2Pin));
      SERIAL_PROTOCOLPGM("E4: ");
      SERIAL_PROTOCOL(   digitalRead(E4ms1Pin));
      SERIAL_PROTOCOLLN( digitalRead(E4ms2Pin));
}

